/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

/* NB: compile with options +l -s */

#include <stdio.h>
#include <exec/memory.h>
#include <intuition/intuition.h>

/* Valid modes for line assembly */
#define _AC 1    /* Accumalator */
#define _IM 2    /* Immediate */
#define _ZP 3    /* Zero page */
#define _ZX 4    /* Zero page X */
#define _ZY 5    /* Zero page Y */
#define _AB 6    /* Absolute */
#define _AX 7    /* Absolute X */
#define _AY 8    /* Absolute Y */
#define _IP 0    /* Implied */
#define _RL 11   /* Relative */
#define _IX 9    /* Indirect X */
#define _IY 10   /* Indirect Y */
#define _AI 12   /* Absolute indirect */

FILE
   *f1, *p1, *fopen(); /* f1 for binary file, p1 for print file */

int
   NBytes[]={
      1, 1, 2, 2, 2, 2, 3, 3, 3, 2, 2, 2, 3 };

/* Mnemonic table */
unsigned long
   Mnemonic[]= {
      'ADC\0','AND\0','ASL\0','BCC\0','BCS\0','BEQ\0',
      'BIT\0','BMI\0','BNE\0','BPL\0','BRK\0','BVC\0',
      'BVS\0','CLC\0','CLD\0','CLI\0','CLV\0','CMP\0',
      'CPX\0','CPY\0','DEC\0','DEX\0','DEY\0','EOR\0',
      'INC\0','INX\0','INY\0','JMP\0','JSR\0','LDA\0',
      'LDX\0','LDY\0','LSR\0','NOP\0','ORA\0','PHA\0',
      'PHP\0','PLA\0','PLP\0','ROL\0','ROR\0','RTI\0',
      'RTS\0','SBC\0','SEC\0','SED\0','SEI\0','STA\0',
      'STX\0','STY\0','TAX\0','TAY\0','TSX\0','TXA\0',
      'TXS\0','TYA\0','???\0' };

/* The table below declares instruction types 0.55 for each possible
 byte 0.255 combination. FE indicates invalid 6502 opcode */
#define FE 56  /* Future expansion! */
unsigned char
   Opcode[]={
      10,34,FE,FE,FE,34, 2,FE, 36,34, 2,FE,FE,34, 2,FE,
       9,34,FE,FE,FE,34, 2,FE, 13,34,FE,FE,FE,34, 2,FE,
      28, 1,FE,FE, 6, 1,39,FE, 38, 1,39,FE, 6, 1,39,FE,
       7, 1,FE,FE,FE, 1,39,FE, 44, 1,FE,FE,FE, 1,39,FE,

      41,23,FE,FE,FE,23,32,FE, 35,23,32,FE,27,23,32,FE,
      11,23,FE,FE,FE,23,32,FE, 15,23,FE,FE,FE,23,32,FE,
      42, 0,FE,FE,FE, 0,40,FE, 37, 0,40,FE,27, 0,40,FE,
      12, 0,FE,FE,FE, 0,40,FE, 46, 0,FE,FE,FE, 0,40,FE,

      FE,47,FE,FE,49,47,48,FE, 22,FE,53,FE,49,47,48,FE,
       3,47,FE,FE,49,47,48,FE, 55,47,54,FE,FE,47,FE,FE,
      31,29,30,FE,31,29,30,FE, 51,29,50,FE,31,29,30,FE,
       4,29,FE,FE,31,29,30,FE, 16,29,52,FE,31,29,30,FE,

      19,17,FE,FE,19,17,20,FE, 26,17,21,FE,19,17,20,FE,
       8,17,FE,FE,FE,17,20,FE, 14,17,FE,FE,FE,17,20,FE,
      18,43,FE,FE,18,43,24,FE, 25,43,33,FE,18,43,24,FE,
       5,43,FE,FE,FE,43,24,FE, 45,43,FE,FE,FE,43,24,FE };
#undef FE
#define FE -1
/* Opcode table */
unsigned char
   Optab[]={
     FE  ,FE  ,0x69,0x65,0x75,FE  ,0x6d,0x7d,0x79,0x61,0x71,FE  ,FE,
     FE  ,FE  ,0x29,0x25,0x35,FE  ,0x2d,0x3d,0x39,0x21,0x31,FE  ,FE,
     FE  ,0x0a,FE  ,0x06,0x16,FE  ,0x0e,0x1e,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0x90,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0xb0,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0xf0,FE,
     FE  ,FE  ,FE  ,0x24,FE  ,FE  ,0x2c,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0x30,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0xd0,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0x10,FE,
     0x00,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0x50,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0x70,FE,
     0x18,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0xd8,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x58,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0xb8,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,0xc9,0xc5,0xd5,FE  ,0xcd,0xdd,0xd9,0xc1,0xd1,FE  ,FE,
     FE  ,FE  ,0xe0,0xe4,FE  ,FE  ,0xec,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,0xc0,0xc4,FE  ,FE  ,0xcc,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,FE  ,0xc6,0xd6,FE  ,0xce,0xde,FE  ,FE  ,FE  ,FE  ,FE,
     0xca,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x88,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,0x49,0x45,0x55,FE  ,0x4d,0x5d,0x59,0x41,0x51,FE  ,FE,
     FE  ,FE  ,FE  ,0xe6,0xf6,FE  ,0xee,0xfe,FE  ,FE  ,FE  ,FE  ,FE,
     0xe8,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0xc8,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0x4c,FE  ,FE  ,FE  ,FE  ,FE  ,0x6c,
     FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,0X20,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,0xa9,0xa5,0xb5,FE  ,0xad,0xbd,0xb9,0xa1,0xb1,FE  ,FE,
     FE  ,FE  ,0xa2,0xa6,FE  ,0xb6,0xae,FE  ,0xbe,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,0xa0,0xa4,0xb4,FE  ,0xac,0xbc,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,0x4a,FE  ,0x46,0x56,FE  ,0x4e,0x5e,FE  ,FE  ,FE  ,FE  ,FE,
     0xea,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,0x09,0x05,0x15,FE  ,0x0d,0x1d,0x19,0x01,0x11,FE  ,FE,
     0x48,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x08,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x68,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x28,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,0x2a,FE  ,0x26,0x36,FE  ,0x2e,0x3e,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,0x6a,FE  ,0x66,0x76,FE  ,0x6e,0x7e,FE  ,FE  ,FE  ,FE  ,FE,
     0x40,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x60,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,0xe9,0xe5,0xf5,FE  ,0xed,0xfd,0xf9,0xe1,0xf1,FE  ,FE,
     0x38,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0xf8,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x78,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,FE  ,0x85,0x95,FE  ,0x8d,0x9d,0x99,0x81,0x91,FE  ,FE,
     FE  ,FE  ,FE  ,0x86,FE  ,0x96,0x8e,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     FE  ,FE  ,FE  ,0x84,0x94,FE  ,0x8c,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0xaa,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0xa8,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0xba,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x8a,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x9a,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE,
     0x98,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE  ,FE };

#define XHOME 4      /* Home coords .x */
#define YHOME 18     /* .y */
#define XWS 328      /* Window size .x */
#define YWS 166      /* .y */
#define XS 8         /* Character width */
#define YS 8         /* Character height */
#define XM 40        /* # columns */
#define YM 19        /* # rows */
#define XB 325       /* Scroll block: right corner .x */
#define YB 163       /* .y */
#define SLEN 79      /* String length */
#define ESC 0x1b     /* Escape code */
#define CR 0x0d      /* Return code */
#define BS 0x08      /* Backspace */
#define SPC 0x20     /* Space */
#define NM1 6        /* Number of menu selections .1 */
#define NM2 14       /* .2 */
#define ILLEGAL 1    /* Illegal opcode for execution */
#define IRQ 2        /* Irq request without vector set up */
#define BREAK 4      /* Break command generates breakpoint */
#define NMI 8        /* Nmi req */

int
   L_on= 0, Quit= FALSE, xp= 0, yp= 0, Dump= FALSE,  XEcute= TRUE, _Di=0, cnt;

unsigned int
   AC= 0, XR= 0, YR= 0, SP= 0xff, PC= 0, SR= 0,  /* 6502 registers */
   MS= 0;   /* MS is internal machine status */

char
   ToAscii(),        /* char functions */
   KEY,
   WTITLE[]="Ape65 v1.0", /* Window title */
   MENU1[]="Info",   /* Menu 1 title */
   MENU2[]="Help",   /* Menu 2 title */
   *SELT1[]= {       /* Menu 1 selections */
      "MCS 6502     ",
      "Simulator &  ",
      "Debugger.    ",
      "-------------",
      "D.M.Broad.(c)",
      " August 1989." },
   *SELT2[]= {       /* Menu 2 selections */
      "A assemble      D disassemble *",
      "C compare *     T transfer     ",
      "F fill          H hunt *       ",
      "G go *          R register dump",
      "M memory dump * > modify memory",
      "; modify reg    I irq *        ",
      "N nmi *         X exit         ",
      "+ trace on      - trace off    ",
      "L load binary   S save binary  ",
      "P list file on  Q list file off",
      "$ hex convert   # dec convert  ",
      "% bin convert   @ oct convert  ",
      "-------------------------------",
      "* Ctrl+p pause  Ctrl+s stop    " },
   _Dstr[SLEN],
   _Tstr[SLEN];

unsigned char
   *Memory=0L;

struct IntuitionBase
   *IntuitionBase=0L;

struct GfxBase
   *GfxBase=0L;

struct NewWindow
   NewWindow= {
   0, 0, XWS, YWS, 0, 1,
   VANILLAKEY | CLOSEWINDOW | MENUPICK,
   WINDOWCLOSE | SMART_REFRESH | ACTIVATE |
   WINDOWDRAG | WINDOWDEPTH | NOCAREREFRESH,
   NULL, NULL,
   WTITLE,
   NULL, NULL,
   0, 0, 0, 0,
   WBENCHSCREEN };

struct Window
   *NWindow=0L;

struct RastPort
   *NRastPort;

struct ViewPort
   *NViewPort;

struct View
   *NView;

struct IntuiText
   DefText= {        /* Default text structure */
   0,1,              /* Used for menus */
   JAM2,
   0,0,
   NULL,NULL,NULL },
   Txt1[NM1],
   Txt2[NM2];

struct MenuItem
   Sel1[NM1],        /* Selection arrays for menus */
   Sel2[NM2];

struct Menu          /* Menu structure .2 */
   Men2= {
   NULL,
   120,0,
   256,10,
   MENUENABLED,
   MENU2,
   Sel2,
   0,0,0,0 },

   Men1= {           /* .1 */
   &Men2,
   0,0,
   120,10,
   MENUENABLED,
   MENU1,
   Sel1,
   0,0,0,0 };

main()
{
   OpenUser();
   do
   {
      GetUser();
   }  while (Quit == FALSE);
   CloseUser();
}

/* Open for business */
OpenUser()
{
   if ( !(Memory= AllocMem (65536, MEMF_PUBLIC | MEMF_CLEAR)) )
      { exit(1); }
   if ( !(IntuitionBase= (struct IntuitionBase *)
      OpenLibrary ("intuition.library", 0L)) )
      { CloseUser(); exit(1); }
   if ( !(GfxBase= (struct GfxBase*)
      OpenLibrary ("graphics.library", 0L)) )
      { CloseUser(); exit(1); }
   if ( !(NWindow=       (struct Window*)
      OpenWindow  (&NewWindow)) )
      { CloseUser(); exit(1); }
   NRastPort= NWindow->RPort;
   NViewPort= ViewPortAddress (NWindow);
   NView= ViewAddress();

   NewMenu (&Men1, SELT1, Sel1, Txt1, NM1, 120);
   NewMenu (&Men2, SELT2, Sel2, Txt2, NM2, 256);
   SetMenuStrip (NWindow, &Men1);
   SetAPen (NRastPort, 1);
   r_c();
   Ps(".");
}

/* Down tools! */
CloseUser()
{
   if (Memory) FreeMem (Memory, 65536);
   if (NWindow) CloseWindow (NWindow);
   if (GfxBase) CloseLibrary (GfxBase);
   if (IntuitionBase) CloseLibrary (IntuitionBase);
}

/* Visit message port */
GetUser()
{
   struct IntuiMessage
      *Mesg;

   WaitPort (NWindow->UserPort);
   if (Mesg= (struct IntuiMessage *)
      GetMsg (NWindow->UserPort))
   {
      ReplyMsg (Mesg);
      switch (Mesg->Class)
      {
         case VANILLAKEY :
            DoChar(Mesg->Code);
            break;
         case CLOSEWINDOW: Quit= TRUE;
            break;
      }
   }
}

/* Visit message port for a key only */
GetKey()
{
   struct IntuiMessage
      *Mesg;

   WaitPort (NWindow->UserPort);
   if (Mesg= (struct IntuiMessage *)
      GetMsg (NWindow->UserPort))
   {
      ReplyMsg (Mesg);
      if (Mesg->Class == VANILLAKEY)
         KEY= (char) Mesg->Code;
      else
         KEY= '\0';
   }
}

/* Add selection structure lists in memory       */
/* Slightly modified NewMenu routine by J.Draper */
NewMenu(menu, name, item, text, num, wid)
struct Menu
   *menu;
struct MenuItem
   item[];
struct IntuiText
   text[];
char
   *name[];
int
   num, wid;
{
   int
      i, hi=0;

   for (i=0; i<num; hi+=10,i++)
   {
      text[i]= DefText;       text[i]. IText= name[i];

      item[i]. NextItem=      &item[i+1];
      item[i]. LeftEdge= 0;   item[i]. TopEdge= 10*i;
      item[i]. Width= wid;    item[i]. Height= 8;
      item[i]. Flags= ITEMTEXT | ITEMENABLED;
      item[i]. MutualExclude= NULL;
      item[i]. ItemFill=   (APTR) &text[i];
      item[i]. SelectFill= (APTR) NULL;
      item[i]. Command= NULL; item[i]. SubItem= NULL;
      item[i]. NextSelect= NULL;
   }
   item[i-1]. NextItem= NULL; menu->Height= hi;
}

   /* No menu support!
DoMenu(MeC)
USHORT
   MeC;
{
   int
      M, S, B, xc, yc;
   char
      st[SLEN];

   M= MENUNUM(MeC); S= ITEMNUM(MeC); B=SUBNUM(MeC);
   switch (M)
   {
      case 0 : Ps("MENU 1 ");
               switch (S)
               {
                  case 0 : Ps("SELECT 1\r"); break;
                  case 1 : Ps("SELECT 2\r"); break;
                  case 2 : Ps("SELECT 3\r"); break;
               }
               break;
      case 1 : Ps("MENU 2 ");
               switch (S)
               {
                  case 0 : Ps("SELECT 1\r"); break;
               }
               break;
   }
}
   */

DoChar(c)      /* Perform function key action */
USHORT
   c;
{
   char
      cc;

   cc= (char) c;
   switch (cc)
   {
      case ESC:
         Quit= TRUE;
         break;
      default :
         UpdateL(&cc);
         break;
   }
}

CurLeft()
{
   if (xp || yp)
   {
      if (xp == 0)
      {
         xp= XM;  /* One too many */
         --yp;
      }
      --xp;
   }
}

UpdateL(ch)
char
   *ch;
{
   static int
      i=0;
   static char
      s[SLEN];
   char
      c;

   if (*ch == BS)
   {
      if (i)
      {
         s[i--]= '\0';
         CurLeft();
         c= SPC;
         UText (&c);
         CurLeft();
      }
   }
   else
   {
      if (i == SLEN)
      {
         Ps("Line too long\r.");
         i= 0;
      }
      else
      {
         UText (ch);
         if (*ch != CR)
         {
            s[i++]= *ch;
            s[i]= '\0';
         }
         else
         {
            if (i)
            {
               PerfFunc(s);
               i= 0;
            }
            Ps(".");
         }
      }
   }
}

itos(s, n, b, w)
char
   s[];
unsigned int
   n;
int
   b, w;
{
   char
      t[SLEN];
   int
      i=0, j=0;

   do
   {
      t[i++]= ToAscii (n % b);
      n= n / b;
   }  while (n != 0);
   while (i < w)
      t[i++]= '0';
   do
   {
      s[j++]= t[--i];
   }  while (i != 0);
   s[j]= '\0';
}

char ToAscii(d)
int
   d;
{
   d+= '0';
   if (d > '9')
      d+= 7;
   return (d);
}

SToUpper(st)
char
   st[];
{
   int
      i=0;

   while (st[i] != '\0')
   {
      st[i]= (char) ToUpper(st[i]);
      i++;
   }
}

ToUpper(c)
char
   c;
{
   if (c >= 'a' && c <= 'z')
      c-= ('a' - 'A');
   return(c);
}

ReadLine(s)
char
   s[];
{
   int
      i=0;
   char
      c= SPC;

   do
   {
      GetKey();
      if ( KEY == BS )
      {
         if (i)
         {
            s[i--]= '\0';
            CurLeft();
            UText(&c);
            CurLeft();
         }
      }
      else
      {
         if (i == SLEN)
         {
            Ps("\rLine too long\r");
            i= 0;
            KEY= CR;
         }
         else
         {
            if (KEY != CR)
            {
               UText(&KEY);
               s[i++]= KEY;
            }
         }
      }
   } while (KEY != CR);
   s[i]= '\0';
}

UText(ch)
char
   *ch;
{
   int
      x, y;

   switch (*ch)
   {
      case CR:
         xp= 0;
         yp++;
         if(L_on) putc('\n', p1);
         break;
      default:
         x= (xp++ * XS) + XHOME; y= (yp * YS) + YHOME;
         if (xp >= XM)
         {
            xp= 0;
            yp++;
         }
         Move (NRastPort, x, y);
         Text (NRastPort, ch, 1);
         if(L_on) putc(*ch, p1);
         break;
   }
   if (yp >= YM)
   {
      ScrollRaster (NRastPort, 0, 8, 3, 12, XB, YB);
      --yp;
   }
}

Ps(s)
char
   s[];
{
   int
      i= 0;

   while (s[i] != NULL)
      UText (&s[i++]);
}

Nl()
{
   char
      k= CR;

   UText (&k);
}

PBWord(w)
unsigned int
   w;
{
   char
      s[SLEN];

   itos(s, w, 2, 16);
   Ps(s);
   Ps(" ");
}

POWord(w)
unsigned int
   w;
{
   char
      s[SLEN];

   itos(s, w, 8, 6);
   Ps(s);
   Ps(" ");
}

PDWord(w)
unsigned int
   w;
{
   char
      s[SLEN];

   itos(s, w, 10, 5);
   Ps(s);
   Ps(" ");
}

PHWord(w)
unsigned int
   w;
{
   char
      s[SLEN];

   itos(s, w, 16, 4);
   Ps(s);
   Ps(" ");
}

PHByte(b)
unsigned int
   b;
{
   char
      s[SLEN];

   itos(s, b, 16, 2);
   Ps(s);
   Ps(" ");
}

Cat(s1, i1, s2)
char
   s1[], s2[];
int
   *i1;
{
   int
      i=0;

   while (s2[i] != '\0')
      s1[(*i1)++]= s2[i++];
}

PerfFunc(s)
char
   s[];
{
   int
      i= 0;
   char
      cm;

   while (s[i] == SPC)
      i++;
   cm= ToUpper(s[i++]);

   switch (cm)
   {
      case 'R': r_c();     /* Register dump */
                break;
      case 'X': Quit= TRUE;/* Exit */
                break;
      case '>': Mod_c(s,i);/* Modify memory */
                break;
      case 'M': m_c(s,i);  /* Memory dump */
                break;
      case 'F': f_c(s,i);  /* Fill memory */
                break;
      case 'T': t_c(s,i);  /* Transfer memory */
                break;
      case 'C': CM_c(s,i); /* Compare memory */
                break;
      case 'H': h_c(s,i);  /* Hunt memory */
                break;
      case ';': SR_c(s,i); /* Set register */
                break;
      case 'G': g_c(s,i);  /* Execute code */
                break;
      case 'I': i_c();     /* IRQ */
                break;
      case 'N': n_c();     /* NMI */
                break;
      case '+': to_c();    /* Trace on */
                break;
      case '-': tf_c();    /* Trace off */
                break;
      case 'D': d_c(s,i);  /* Disassemble */
                break;
      case 'A': a_c(s,i);  /* Assemble */
                break;
      case 'S': S_c(s,i);  /* Save */
                break;
      case 'L': L_c(s,i);  /* Load */
                break;
      case 'P': P_c(s,i);  /* Print/List file open */
                break;
      case 'Q': Q_c();     /*  "     "   file close */
                break;
      case '#': D_c(s,i);  /* Base conversions */
                break;
      case '$': H_c(s,i);
                break;
      case '@': O_c(s,i);
                break;
      case '%': B_c(s,i);
                break;
      default:  Ps("Unknown command");
                Nl();
                break;
   }
}

r_c()
{
   int
      i, m=128;

   Ps("  AC XR YR SR SP  PC  NV.BDIZC");
   Nl();
   Ps("  ");
   PHByte(AC);
   PHByte(XR);
   PHByte(YR);
   PHByte(SR);
   PHByte(SP);
   PHWord(PC);
   while (m != 0)
   {
      if (SR & m)
         Ps("1");
      else
         Ps(".");
      m= m / 2;
   }
   Nl();
}

P_c(s, i)
char
   s[];
int
   i;
{
   if(L_on) { Ps("List file already active\r"); return(); }
   _Tstr[0]='\0';
   GetF(s, &i, _Tstr);
   if( (p1= fopen(_Tstr, "w")) == NULL)
   { Ps("Can't open list file\r"); return();}
   L_on= 1; /* Flag to write to list */
}

Q_c()
{
   if(!L_on) { Ps("No list file active\r"); return(); }
   L_on= 0;
   fclose(p1);
}

S_c(s, i)
char
   s[];
int
   i;
{
   unsigned int
      f=0, l=0, si;

   _Tstr[0]='\0';
   GetF(s, &i, _Tstr);
   GetInt(s, &i, &f, 16);
   GetInt(s, &i, &l, 16);
   if( (_Tstr[0]=='\0') || (f==l) || (l<f) ) return();

   if( (f1= fopen(_Tstr, "w")) == NULL)
   { Ps("Can't open output file\r"); return();}
   putc( f & 0xff, f1);         /* write low byte address */
   putc( (f >> 8) & 0xff, f1);  /* write high byte address */
   si=l-f;
   putc( si & 0xff, f1);        /* write low byte length */
   putc( (si >> 8) & 0xff, f1); /* write high byte length */

   while (f < l)
      putc( *(Memory + f++), f1);
   fclose(f1);
}

L_c(s, i)
char
   s[];
int
   i;
{
   unsigned int
      f=0, l=0, s1, s2;

   _Tstr[0]='\0';
   GetF(s, &i, _Tstr);
   GetInt(s, &i, &f, 16);
   if(_Tstr[0]=='\0') return();

   if( (f1= fopen(_Tstr, "r")) == NULL)
   { Ps("Can't open input file\r"); return();}
   s1=getc(f1);         /* read low byte address */
   s2=getc(f1);         /* read high byte address */
   if (!f) f= (s2 << 8) | s1;
   Ps("Load start: "); PHWord(f=f & 0xffff); Nl();
   s1=getc(f1);         /* read low byte length */
   s2=getc(f1);         /* read high byte length */
   l= (((s2 << 8) | s1) + f) & 0xffff;
   while (f < l)
      *(Memory + f++)= getc(f1);
   fclose(f1);
   Ps("Load end:   "); PHWord(f); Nl();
}

GetF(s1, i1, s2)
char
   s1[], s2[];
int
   *i1;
{
   int
      i2=0, qf=0, gl=1;

   SkipBlank(s1, i1);
   if (s1[*i1]=='"')
   {
      qf=1; (*i1)++;
   }
   while(gl)
   {
   if ( (qf && (s1[*i1]== '"')) || (!qf && (s1[*i1]== SPC)) ||
         (s1[*i1]== '\0') ) gl=0;
   else
      s2[ i2++ ] = s1[ (*i1)++ ];
   }
   if (qf && (s1[*i1]=='"')) (*i1)++;
   s2[i2]='\0';
}

m_c(s, i)
char
   s[];
int
   i;
{
   static unsigned int
      f= 0;
   unsigned int
      d, flg= FALSE, l= 0, e;
   char
      db[9];

   db[9]='\0';
   GetInt(s, &i, &f, 16);
   GetInt(s, &i, &l, 16);
   if (f > l)
      l= f;
   cnt=MS= 0;
   while (flg == FALSE && MS == 0)
   {
      Ps("  "); PHWord(f);
      for (d=0; d<8; d++)
      {
         if (f == l)
            flg= TRUE;
         e= *(Memory + f++);  /* Peek */
         f= f & 0xffff;
         PHByte(e);
         if (e < SPC)
            e= '.';
         db[d]= e;
      }
      db[8]='\r'; Ps(db);
      if (++cnt == 18)
      {
         Ps("...");
         GetKey();
         KEY=xp=cnt= 0;
      }
      ChkPort();
   }
}

D_c(s, i)
char
   s[];
int
   i;
{
   unsigned int
      n=0;

   GetInt(s, &i, &n, 10);
   Conv(&n);
}

H_c(s, i)
char
   s[];
int
   i;
{
   unsigned int
      n=0;

   GetInt(s, &i, &n, 16);
   Conv(&n);
}

B_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      n=0;

   GetInt(s, &i, &n, 2);
   Conv(&n);
}

O_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      n=0;

   GetInt(s, &i, &n, 8);
   Conv(&n);
}

Conv(n)
unsigned int
   *n;
{
   if (*n)
   {
      Ps("%"); PBWord(*n);
      Ps("@"); POWord(*n);
      Ps("#"); PDWord(*n);
      Ps("$"); PHWord(*n);
      Nl();
   }
}

GetInt(st, in, as, b)
char
   st[];
int
   *in, b;
unsigned int
   *as;
{
   long
      i= 0x10000;

   while ((st[*in] == SPC) || (st[*in] == ','))
      (*in)++;
   while (  (st[*in] >= 'a') && (st[*in] <= 'z')
         || (st[*in] >= 'A') && (st[*in] <= 'Z')
         || (st[*in] >= '0') && (st[*in] <= '9')  )
      i= i * b + ToNum(st[(*in)++], b);
   if (i != 0x10000)
      *as= (unsigned int) i & 0xffff;
}

ToNum(c, ba)
char
   c;
int
   ba;
{
   c= ToUpper(c);
   if (c >= 'A' && c <= 'Z')
      c-= ('A' - 10);
   else if (c >= '0' && c <= '9')
      c-= '0';
   else
      c= 0;
   if (c >= ba)
      c= 0;
   return (c);
}

Mod_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      ad= 0, m;

   GetInt(s, &i, &ad, 16);
   while ((s[i] == SPC) || (s[i] == ','))
      i++;
   if (s[i] == '\0') return();
   if ((s[i] == '"') || (s[i] == '\''))
   {
      i++;  /* Skip quote */
      while (s[i] != '\0')
      {
         *(Memory + ad++)= s[i++];
         ad= ad & 0xffff;
      }
   }
   else
   {
      while (s[i] != '\0')
      {
         GetInt(s, &i, &m, 16);
         *(Memory + ad++)= m;    /* Poke */
         ad= ad & 0xffff;
      }
   }
}

f_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      f= 0, l= 0, w= 0;

   GetInt(s, &i, &f, 16);
   GetInt(s, &i, &l, 16);
   GetInt(s, &i, &w, 16);
   while (f <= l)
      *(Memory + f++)= w;     /* Poke */
}

t_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      a1= 0, e1= 0, a2= 0, e2=0;

   GetInt(s, &i, &a1, 16);
   GetInt(s, &i, &e1, 16);
   GetInt(s, &i, &a2, 16);
   if (e1 < a1)
      return();
   if (a1 == a2)
      return();
   e2= (a2 + (e1 - a1)) & 0xffff;
   /* Perform transfer in direction so as not to cause
      overlay fault. Can't use '<=' or '>=' in loops because
      argument may inc or dec to 0000 or ffff.. Eg T0,0F,20 */
   if (a1 > a2)
   {
      while (a1 != e1)
      {
         *(Memory + a2++)= *(Memory + a1++);
         a2= a2 & 0xffff; a1= a1 & 0xffff;
      }
      *(Memory + a2)= *(Memory + a1); /* Copy last byte */
   }
   else
   {
      while (e1 != a1)
      {
         *(Memory + e2--)= *(Memory + e1--);
         e2= e2 & 0xffff; e1= e1 & 0xffff;
      }
      *(Memory + e2)= *(Memory + e1);  /* Copy last byte */
   }
}

CM_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      a1= 0, e1= 0, a2= 0, m= 0;

   GetInt(s, &i, &a1, 16);
   GetInt(s, &i, &e1, 16);
   GetInt(s, &i, &a2, 16);
   if (e1 < a1)
      return();
   if (a1 == a2)
      return();
   MS= 0;
   while (a1 != e1 && MS == 0)
   {
      if (*(Memory + a1++) != *(Memory + a2++))
      {
         PHWord(a1 - 1); m++;
      }
      a2= a2 & 0xffff; a1= a1 & 0xffff;
      ChkPort();
   }
   if (*(Memory + a1) != *(Memory + a2))
   {
      PHWord(a1); m++;
   }
   if (m % 8)  /* There's eight words a line printed */
      Nl();    /* Do a newline if cursor left astray */
}

h_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      ht[SLEN],   /* Hunt parameters */
      a1, e1, ix= 0, m= 0, ha, hi;

   GetInt(s, &i, &a1, 16);
   GetInt(s, &i, &e1, 16);
   if (e1 < a1)
      return();
   while ((s[i] == SPC) || (s[i] == ','))
      i++;
   if (s[i] == '\0') return();
   if ((s[i] == '"') || (s[i] == '\''))
   {
      i++;  /* Skip quote */
      while ((s[i] != '\0') && (ix < SLEN))
         ht[ix++]= s[i++];
   }
   else
   {
      while ((s[i] != '\0') && (ix < SLEN))
         GetInt(s, &i, &(ht[ix++]), 16);
   }
   MS= 0;
   while (a1 != e1 && MS == 0)
   {
      hi= 0; ha= a1;
      while ((*(Memory + ha++) == ht[hi]) && (hi < ix))
      {
         hi++; ha= ha & 0xffff;
      }
      if (hi >= ix)
      {
         PHWord(a1); m++;
      }
      a1= (a1 + 1) & 0xffff;
      ChkPort();
   }
   if (m % 8)
      Nl();
}

SR_c(s,i)
char
   s[];
int
   i[];
{
   GetInt(s, &i, &AC, 16); AC= AC & 0xff;
   GetInt(s, &i, &XR, 16); XR= XR & 0xff;
   GetInt(s, &i, &YR, 16); YR= YR & 0xff;
   GetInt(s, &i, &SR, 16); SR= SR & 0xff;
   GetInt(s, &i, &SP, 16); SP= SP & 0xff;
   GetInt(s, &i, &PC, 16); PC= PC & 0xffff;
}

Do6502() /* Execute opcode */
{
   unsigned int
      op, opc, a1=0, a2=0, ea, b1;
   char
      *c, k;

   /* opc is instr type op is opcode */
   opc= Opcode[ op=*(Memory + PC) ];
   if (Dump)
   {
      _Di= 0;
      itos(_Tstr, PC, 16, 4);
      Cat(_Dstr, &_Di, _Tstr); _Dstr[_Di++]= SPC;
      c= (char*) &Mnemonic[opc];
      Cat(_Dstr, &_Di, c);     _Dstr[_Di++]= SPC;
   }
   PC= (PC + 1) & 0xffff;

   GAddr(&ea, op);

   if (XEcute) switch (opc)
   {
      case 0 :
            /* ADC */
            SR= SR & 0xbf;
            b1= *(Memory + ea);
            a1= AC; a2= b1;
            Flag( (b1=AC= AC + (SR & 1) + b1) );
            if ( 128 &
            ((a1 & a2 & (b1 ^ 0xff)) ^ ((a1 ^ 0xff) & (a2 ^ 0xff) &b1)) )
               SR= SR | 64;
            if (SR & 8)
            {
               if ((AC & 0x0f) > 9)
                  AC= AC + 6;
               if ((AC & 0xf0) > 0x90)
                  AC= AC + 0x60;
            }
            SR= (SR & 0xfe) | (AC >> 8);
            AC= AC & 0xff;
            break;
      case 1 :
            /* AND */
            Flag( (AC= AC & *(Memory + ea)) );
            break;
      case 2 :
            /* ASL */
            if (op == 0x0a)
            {
               b1= AC << 1;
               Flag( (AC= b1 & 0xfe) );
            }
            else
            {
               b1= *(Memory + ea) << 1;
               Flag( (*(Memory + ea)= b1 & 0xfe) );
            }
            SR= (SR & 0xfe) | (b1 >> 8);
            break;
      case 3 :
            if (!(SR & 1)) PC= ea; /* BCC */
            break;
      case 4 :
            if (SR & 1) PC= ea;    /* BCS */
            break;
      case 5 :
            if (SR & 2) PC= ea;    /* BEQ */
            break;
      case 6 :
            /* BIT */
            b1= *(Memory + ea);
            SR= (SR & 0x3d) | (b1 & 0xc0);
            if (!(AC & b1))
               SR= SR | 2;
            break;
      case 7 :
            if (SR & 128) PC= ea;  /* BMI */
            break;
      case 8 :
            if (!(SR & 2)) PC= ea; /* BNE */
            break;
      case 9 :
            if (!(SR & 128))PC= ea;/* BPL */
            break;
      case 10:
            GNext(&a1);    /* BRK */
            PStack(PC >> 8);
            PStack(PC & 0xff);
            PStack( SR | 16 );
            MS= MS | BREAK;
            break;
      case 11:
            if (!(SR & 64)) PC=ea; /* BVC */
            break;
      case 12:
            if (SR & 64) PC= ea;   /* BVS */
            break;
      case 13:
            SR= SR & 0xfe;    /* CLC */
            break;
      case 14:
            SR= SR & 0xf7;    /* CLD */
            break;
      case 15:
            SR= SR & 0xfb;    /* CLI */
            break;
      case 16:
            SR= SR & 0xbf;    /* CLV */
            break;
      case 17:
            /* CMP */
            a1= *(Memory + ea);
            Flag( (b1= (AC - a1) & 0xff) );
            SR= SR & 0xfe;
            if (AC >= a1) SR= SR | 1;
            break;
      case 18:
            /* CPX */
            a1= *(Memory + ea);
            Flag( (b1= (XR - a1) & 0xff) );
            SR= SR & 0xfe;
            if (XR >= a1) SR= SR | 1;
            break;
      case 19:
            /* CPY */
            a1= *(Memory + ea);
            Flag( (b1= (YR - a1) & 0xff) );
            SR= SR & 0xfe;
            if (YR >= a1) SR= SR | 1;
            break;
      case 20:
            /* DEC */
            Flag( (b1= *(Memory + ea) - 1) );
            *(Memory + ea)= b1;
            break;
      case 21:
            Flag( (XR= (XR - 1) & 0xff) ); /* DEX */
            break;
      case 22:
            Flag( (YR= (YR - 1) & 0xff) ); /* DEY */
            break;
      case 23:
            /* EOR */
            Flag( (AC= (AC ^ *(Memory + ea)) & 0xff) );
            break;
      case 24:
            /* INC */
            Flag( (b1= *(Memory + ea) + 1) );
            *(Memory + ea)= b1;
            break;
      case 25:
            Flag( (XR= (XR + 1) & 0xff) ); /* INX */
            break;
      case 26:
            Flag( (YR= (YR + 1) & 0xff) ); /* INY */
            break;
      case 27:
            /* JMP */
            if (XEcute) PC= ea;
            break;
      case 28:
            /* JSR */
            switch(ea)
            {
               case 65490:  /* Print character $FFD2 */
                  k= (char) AC;
                  UText(&k);
                  break;
               case 65508:  /* Read character  $FFE4 */
                  Flag( AC= KEY & 0xff );
                  KEY= '\0'; /* Ok.. ready for another */
                  break;
               default:
                  b1= (PC - 1) & 0xffff;
                  PStack(b1 >> 8);
                  PStack(b1 & 0xff);
                  PC= ea;
                  break;
            }
            break;
      case 29:
            /* LDA */
            Flag( (AC= *(Memory + ea)) );
            break;
      case 30:
            /* LDX */
            Flag( (XR= *(Memory + ea)) );
            break;
      case 31:
            /* LDY */
            Flag( (YR= *(Memory + ea)) );
            break;
      case 32:
            if (op == 0x4a)      /* LSR */
            {
               b1= AC >> 1; a1= AC & 1;
               Flag( (AC= b1 & 0x7f) );
            }
            else
            {
               b1= *(Memory + ea) >> 1; a1= *(Memory + ea) & 1;
               Flag( (*(Memory + ea)= b1 & 0x7f) );
            }
            SR= (SR & 0xfe) | a1;
            break;
      case 33:                /* NOP */
            break;
      case 34:
            /* ORA */
            Flag( (AC= AC | *(Memory + ea)) );
            break;
      case 35:
            PStack(AC);          /* PHA */
            break;
      case 36:
            PStack(SR);          /* PHP */
            break;
      case 37:
            Flag( (GStack(&AC)) );         /* PLA */
            break;
      case 38:
            GStack(&SR);         /* PLP */
            break;
      case 39:
            if (op == 0x2a)      /* ROL */
            {
               b1= AC << 1;
               Flag( (AC= (b1 & 0xfe) | (SR & 1)) );
            }
            else
            {
               b1= *(Memory + ea) << 1;
               Flag( (*(Memory + ea)= (b1 & 0xfe) | (SR & 1)) );
            }
            SR= (SR & 0xfe) | (b1 >> 8);
            break;
      case 40:
            if (op == 0x6a)     /* ROR */
            {
               b1= AC >> 1; a1= AC & 1;
               Flag( (AC= ((b1 & 0x7f) | (SR << 7)) & 0xff) );
            }
            else
            {
               b1= *(Memory + ea) >> 1; a1= *(Memory + ea) & 1;
               Flag( (*(Memory + ea)= ((b1 & 0x7f) | (SR << 7)) & 0xff) );
            }
            SR= (SR & 0xfe) | a1;
            break;
      case 41:
            GStack(&SR);         /* RTI */
            GStack(&a1);
            GStack(&a2);
            PC= (a2 << 8) | a1;
            SR= SR & 0xeb; /* CLIflag */
            MS= MS | BREAK;
            break;
      case 42:
            GStack(&a1);         /* RTS */
            GStack(&a2);
            PC= (((a2 << 8) | a1) + 1) & 0xffff;
            break;
      case 43: /* SBC */
            SR= SR & 0xbf;
            b1= *(Memory + ea) ^ 0xff;
            a1= AC; a2= b1;
            Flag( (b1=AC= AC + (SR & 1) + b1) );
            if ( 128 &
            ((a1 & a2 & (b1 ^ 0xff)) ^ ((a1 ^ 0xff) & (a2 ^ 0xff) &b1)) )
               SR= SR | 64;
            if (SR & 8)
            {
               if ((AC & 0x0f) > 9)
                  AC= AC + 6;
               if ((AC & 0xf0) > 0x90)
                  AC= AC + 0x60;
            }
            SR= (SR & 0xfe) | (AC >> 8);
            AC= AC & 0xff;
            break;
      case 44:
            SR= SR | 1;          /* SEC */
            break;
      case 45:
            SR= SR | 8;          /* SED */
            break;
      case 46:
            SR= SR | 4;          /* SEI */
            break;
      case 47:
            /* STA */
            *(Memory + ea)= AC;
            break;
      case 48:
            /* STX */
            *(Memory + ea)= XR;
            break;
      case 49:
            /* STY */
            *(Memory + ea)= YR;
            break;
      case 50:
            Flag( (XR= AC) );              /* TAX */
            break;
      case 51:
            Flag( (YR= AC) );              /* TAY */
            break;
      case 52:
            Flag( (XR= SP) );              /* TSX */
            break;
      case 53:
            Flag( (AC= XR) );              /* TXA */
            break;
      case 54:
            SP= XR;              /* TXS */
            break;
      case 55:
            Flag( (AC= YR) );              /* TYA */
            break;
      case 56:
            MS= MS | ILLEGAL;
            break;
   }
   if (Dump && !XEcute)
   {
      Ps("  ");
      _Dstr[_Di]= '\0'; Ps(_Dstr); Nl();
      if (++cnt == 18)
      {
         Ps("...");
         GetKey();
         KEY=xp=cnt= 0;
      }
   }
   if (Dump && XEcute)
   {
      _Dstr[_Di]= '\0'; Ps(_Dstr); Nl();
      r_c();
      if (++cnt == 6)
      {
         Ps("...");
         GetKey();
         KEY=xp=cnt= 0;
      }
   }
   ChkPort();
}

g_c(s,i)
char
   s[];
int
   i;
{
   unsigned int
      ea;

   cnt=MS=0; /* Internal status OK */
   ea= PC;
   GetInt(s, &i, &ea, 16);
   PC= ea;
   while (MS == 0)
      Do6502();
   if ( !(Dump && XEcute) )
      r_c();
}

PStack(b)
unsigned int
   b;
{
   *(Memory + 256 + SP)= (unsigned char) b;
   SP= (SP - 1) & 0xff;
}

GStack(b)
unsigned int
   *b;
{
   SP= (SP + 1) & 0xff;
   *b= *(Memory + 256 + SP);
}

GNext(b)
unsigned int
   *b;
{
   *b= *(Memory + PC);
   PC= (PC + 1) & 0xffff;
}


Flag(a)
unsigned int
   a;
{
   SR= (SR & 0x7d) | (a & 0x80); /* Reset zero, test sign */
   if (!(a & 0xff))
      SR= SR | 2; /* Set zero flag if a = 0 */
}


x_IRQ()
{
   unsigned int
      i, j=65534, k=65535;

   i= *(Memory + j) + (*(Memory + k) << 8);
   if (i == 0)
   {
      MS= MS | IRQ; /* Exit with Irq flag */
      Ps("  IRQ vector not initialised\r");
   }
   else
   {
      Ps("  IRQ service\r");
      PStack(PC >> 8);
      PStack(PC & 0xff);
      PStack(SR);
      SR= SR | 4;
      PC= i;
   }
}

i_c()
{
   if (SR & 4)
      Ps("  Interrupt disabled\r");
   else
   {
      cnt=MS= 0;
      x_IRQ();
      if (MS == 0)
      {
         while (MS == 0)
            Do6502();
         if ( !(Dump && XEcute) ) r_c();
         SR= SR & 0xfb; /* reset irq flag */
      }
   }
}

x_NMI()
{
   unsigned int
      i, j=65530, k=65531;

   i= *(Memory + j) + (*(Memory + k) << 8);
   if (i == 0)
   {
      MS= MS | NMI; /* Exit with NMI flag */
      Ps("  NMI vector not initialised\r");
   }
   else
   {
      Ps("  NMI service\r");
      PStack(PC >> 8);
      PStack(PC & 0xff);
      PStack(SR);
      SR= SR | 4;
      PC= i;
   }
}

n_c()
{
   unsigned int
      t;

   cnt=MS= 0;
   t= SR & 4;
   x_NMI();
   if (MS == 0)
   {
      while (MS == 0)
         Do6502();
      if ( !(Dump && XEcute) ) r_c();
      SR= (SR & 0xfb) | t;
   }
}

to_c()
{
   Ps("Trace ON\r");
   Dump= TRUE;
}

tf_c()
{
   Ps("Trace OFF\r");
   Dump= FALSE;
}

d_c(s,i)
char
   s[];
int
   i;
{
   static unsigned int
      f;
   unsigned int
      l=0, t1, t2;

   GetInt(s, &i, &f, 16);
   GetInt(s, &i, &l, 16);
   if (f >= l)
      l= (f+1) & 0xffff;
   XEcute= FALSE; t1= Dump; t2= PC;
   Dump=TRUE;
   PC=f;
   cnt=MS=0;
   while (PC < l && MS == 0)
      Do6502();
   f=PC; /* Update static variable */
   XEcute= TRUE;  Dump=t1;  PC=t2;
}

a_c(s,i)
char
   s[];
int
   i;
{
   char
      tt[SLEN], st[SLEN], ch;
   unsigned int
      ad;
   static unsigned int
      f;
   int
      n, o, p, m;

   GetInt(s, &i, &f, 16);
   do
   {
      Ps("..");
      itos(st, f, 16, 4);
      Ps(st);
      Ps(" ");
      ReadLine(st);
      SToUpper(st);
      p= Search((st[0] << 24) | (st[1] << 16) | (st[2] << 8));
      if (p != -1)
      {
         ad= f;
         m= GMode(st, p, &ad);   /* Get mode and data */
         o= GCode(p, &m);        /* Get opcode */
         if (o == 0xff)
            p= -1;   /* Syntax error! */
         else
         {
            n= NBytes[m]; /* Get # of bytes */
            GenBytes(tt, o, ad, n);
            StoBytes(&f, o, ad, n);
            xp= 7;
            if (L_on) putc('\n', p1);
            Ps(tt);
            Ps(st);
         }
      }
      Nl();
   }  while (p != -1);
}

Search(M)
unsigned long
   M;
{
   int
      x=0;

   while ((Mnemonic[x] != M) && (x < 56))
      x++;
   if (x < 56)
      return(x);
   else
      return(-1);
}

GMode(s, m, a)
char
   s[];
unsigned int
   *a, m;
{
   int
      i=3;
   unsigned int
      b;

   SkipBlank(s, &i);
   if (s[i] == '\0')
      return (_IP);
   if (s[i] == '#')
   {
      i++;
      GetInt(s, &i, a, 16);
      return(_IM);
   }
   if (s[i] == '(')
   {
      i++;
      GetInt(s, &i, a, 16);
      if (s[i] == ',' && s[i+1] == 'X' && s[i+2] == ')')
         return(_IX);
      if (s[i] == ')' && s[i+1] == ',' && s[i+2] == 'Y')
         return(_IY);
      return(_AI);
   }
   if (s[i] == 'A' && s[i+1] == '\0')
       return(_AC);
   b= (*a + 2) & 0xffff;
   GetInt(s, &i, a, 16);
   if (m >= 3 && m <= 12 && m != 6 && m != 10)
   {
      if (*a >= b)
      {  /* Forward branch */
         if ((*a - b) > 127)
            return(_AI);/* Force error */
         else
            *a= *a - b;
      }
      else
      {
         if ((b - *a) > 128)
            return(_AI);
         else
            *a= ((b - *a) ^ 0xff) + 1;
      }
      return(_RL);
   }
   if (s[i] == '\0')
   {
      if (*a > 0xff)
         return(_AB);
      else
         return(_ZP);
   }
   if (s[i] == ',')
   {
      if (s[i+1] == 'X')
      {
         if (*a > 0xff)
            return(_AX);
         else
            return(_ZX);
      }
      if (s[i+1] == 'Y')
      {
         if (*a > 0xff)
            return(_AY);
         else
            return(_ZY);
      }
   }
   return(0);
}

SkipBlank(st, ix)
char
   st[];
int
   *ix;
{
   while (st[*ix] == SPC)
      (*ix)++;
}

GCode(op, md)
int
   op, *md;
{
   int
      w;

   w= Optab[ (op * 13) + *md ];
   if (w == 0xff)
   {
      if (*md == _ZX) *md= _AX;
      if (*md == _ZY) *md= _AY;
      if (*md == _ZP) *md= _AB;
      w= Optab[ (op * 13) + *md ];
   }
   return(w);
}

GenBytes(ss, oo, aa, nn)
char
   ss[];
int
   oo, aa, nn;
{
   char
      ts[8];
   int
      i;

   ss[9]= '\0';
   for (i=0; i<9; i++) ss[i]= SPC;
   itos(ts, oo, 16, 2);
   ss[0]= ts[0]; ss[1]= ts[1];
   if (nn > 1)
   {
      itos(ts, aa, 16, 4);
      ss[3]= ts[2]; ss[4]= ts[3];
      if (nn == 3)
      {
         ss[6]= ts[0]; ss[7]= ts[1];
      }
   }
}

StoBytes(ff, oo, aa, nn)
unsigned int
   *ff;
int
   oo, aa, nn;
{
   *(Memory + *ff)= oo & 0xff;
   *ff= (*ff + 1) & 0xffff;
   if (nn > 1)
   {
      *(Memory + *ff)= aa & 0xff;
      *ff= (*ff + 1) & 0xffff;
      if (nn == 3)
      {
         *(Memory + *ff)= (aa >> 8) & 0xff;
         *ff= (*ff + 1) & 0xffff;
      }
   }
}


ChkPort()
{
   struct IntuiMessage
      *Mesg;
   char
      c;

   if (Mesg= (struct IntuiMessage *)
      GetMsg(NWindow->UserPort))
   {
      ReplyMsg(Mesg);
      if (Mesg->Class == VANILLAKEY)
      {
         switch (Mesg->Code)
         {
            case 16 : GetKey();        /* Ctrl+p pause */
                      break;
            case 19 : MS= MS | BREAK;  /* Ctrl+s stop */
                      break;
            default : KEY= (char) Mesg->Code;
                      break;
         }
      }
   }
}

GAddr(ea, c)   /* Return address of data/ only for relevant commands */
unsigned int
   *ea, c;
{
   unsigned int
      i=0, im, t=16, nb, a1, a2, b, ad;

   im= ( 13 * (Opcode[c]) );
   do
   {
      if (Optab[im + i] == c)
      {
         t= i;
         break;
      }
      i++;
   }  while (i < 13);

   nb= NBytes[t];
   if (nb > 1)
   {
      GNext(&a1);
      if (nb == 3) GNext(&a2);
   }

   if (XEcute) switch(t)
   {
      case 0:  break;                        /* IMP */
      case 1:  break;                        /* A */
      case 2 : *ea= PC - 1;                  /* # */
               break;
      case 3 : *ea= a1;                      /* ZP */
               break;
      case 4 : *ea= a1 + XR;                 /* ZP x */
               break;
      case 5 : *ea= a1 + YR;                 /* ZP y */
               break;
      case 6 : *ea= (a2 << 8) | a1;          /* ABS */
               break;
      case 7 : *ea= ((a2 << 8) | a1) + XR;   /* ABS x */
               break;
      case 8 : *ea= ((a2 << 8) | a1) + YR;   /* ABS y */
               break;
      case 9 : b= *(Memory + XR + a1);     /* IX */
               *ea= (*(Memory + XR + (a1 + 1)) << 8) | b;
               break;
      case 10: b= *(Memory + a1);          /* IY */
               *ea= ((*(Memory + (a1 + 1)) << 8) | b) + YR;
               break;
      case 11: if (a1 < 128)                 /* REL */
                  *ea= a2= (PC + a1) & 0xffff;
               else
                  *ea= a2= (PC - 256 + a1) & 0xffff;
               break;
      case 12: a1= (a2 << 8) | a1;           /* AI */
               a2= *(Memory + a1);
               *ea= a2 | (*(Memory + a1) << 8);
      default: break;
   }
   *ea= (*ea) & 0xffff;

   if (Dump)
   {
   /* Disassemble or trace only... */
   if (nb == 2) itos(_Tstr, a1, 16, 2);
   if (nb == 3) itos(_Tstr, (a2 << 8) | a1, 16, 4);

   switch(t)
   {
      case 0 : break;
      case 1 : _Dstr[_Di++]= 'A';
               break;
      case 2 : _Dstr[_Di++]= '#';
               Cat(_Dstr, &_Di, _Tstr);
               break;
      case 3 : Cat(_Dstr, &_Di, _Tstr);
               break;
      case 4 : Cat(_Dstr, &_Di, _Tstr);
               Cat(_Dstr, &_Di, ",X");
               break;
      case 5 : Cat(_Dstr, &_Di, _Tstr);
               Cat(_Dstr, &_Di, ",Y");
               break;
      case 6 : Cat(_Dstr, &_Di, _Tstr);
               break;
      case 7 : Cat(_Dstr, &_Di, _Tstr);
               Cat(_Dstr, &_Di, ",X");
               break;
      case 8 : Cat(_Dstr, &_Di, _Tstr);
               Cat(_Dstr, &_Di, ",Y");
               break;
      case 9 : _Dstr[_Di++]= '(';
               Cat(_Dstr, &_Di, _Tstr);
               Cat(_Dstr, &_Di, ",X)");
               break;
      case 10: _Dstr[_Di++]= '(';
               Cat(_Dstr, &_Di, _Tstr);
               Cat(_Dstr, &_Di, "),Y");
               break;
      case 11: if (a1 < 128)
                  a2= (PC + a1) & 0xffff;
               else
                  a2= (PC - 256 + a1) & 0xffff;
               itos(_Tstr, a2, 16, 4);
               Cat(_Dstr, &_Di, _Tstr);
               break;
      case 12: _Dstr[_Di++]= '(';
               Cat(_Dstr, &_Di, _Tstr);
               _Dstr[_Di++]= ')';
      default: break;
   }
   }
}

