/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

/*
 * 6502 simulator
 */

#include "ape65.h"

uint8_t *memory = NULL;

extern const uint8_t BASIC[BASIC_SIZE];

extern const uint8_t MONITOR[MONITOR_SIZE];

uint8_t
peek(uint16_t address)
{
	if (address < RAM_SIZE) {
		return memory[address];
        }
	if (address >= BASIC_ADDR && address < (BASIC_ADDR + BASIC_SIZE)) {
		return BASIC[address - BASIC_ADDR];
	}
	if (address >= MONITOR_ADDR) {
		return MONITOR[address - MONITOR_ADDR];
	}
	if (address == 0xF004) {
		return getcharacter();
	}
	if (address == 0xFD10) {
		uint8_t value;

		value =  (LATAbits.LATA0  == 0) ? 1 : 0; /* LED1  O/P STATE */
		value |= (LATBbits.LATB15 == 0) ? 2 : 0; /* LED2  O/P STATE */
		value |= (LATAbits.LATA4  == 0) ? 4 : 0; /* LEDRX O/P STATE */
		value |= (LATBbits.LATB4  == 0) ? 8 : 0; /* LEDTX O/P STATE */

		value |= (PORTBbits.RB5  == 1) ? 16 : 0; /* D2 INPUT */
		value |= (PORTBbits.RB13 == 1) ? 32 : 0; /* D3 INPUT */
		value |= (PORTBbits.RB10 == 1) ? 64 : 0; /* D8 INPUT */

		value |= (PORTBbits.RB9 == 0) ? 128 : 0; /* SW: BOOTLOAD-EN I/P */
		return value;
	}
	return 0xFF;
}

void
poke(uint16_t address, uint8_t value)
{
	if (address < RAM_SIZE) {
		memory[address] = value;
		return;
	}
	if (address == 0xF001) {
		putcharacter(value);
		return;
	}
	if (address == 0xFD10) {
		LATAbits.LATA0  = !(value & 1);	/* LED1  O/P */
		LATBbits.LATB15 = !(value & 2);	/* LED2  O/P */
		LATAbits.LATA4  = !(value & 4); /* LEDRX O/P */
		LATBbits.LATB4  = !(value & 8); /* LEDTX O/P */

		TRISBbits.TRISB5  = (value & 16) != 0; /* D2 */
		TRISBbits.TRISB13 = (value & 32) != 0; /* D3 */
		TRISBbits.TRISB10 = (value & 64) != 0; /* D8 */
		return;
	}
}

static inline void
init_pps(void)
{
	/* TX2 = RB0(4) RX2 = RB1(5) */
	PPSUnLock;
	PPSOutput(4, RPB0, U2TX);
	PPSInput(2, U2RX, RPB1);
	PPSLock;
}

static inline void
init_io(void)
{
	/* ADC Digital Mode */
	ANSELA = 0;
	ANSELB = 0;

	/*
	 * PORT A/B
	 */
	
	LATA  = 0x0011;		/* LEDs OFF       */
	TRISA = 0xFFFF;		/* DEFAULT TO I/P */

	TRISAbits.TRISA0 = 0;	/* LED1   O/P     */
	TRISAbits.TRISA4 = 0;	/* LEDRX1 O/P     */

	LATB  = 0x8010;		/* LEDs OFF       */
	TRISB = 0xFFFF;		/* DEFAULT TO I/P */

	TRISBbits.TRISB15 = 0;	/* LED2   O/P     */
	TRISBbits.TRISB4  = 0;	/* LEDTX1 O/P     */
}

int
main(int argc, char **argv)
{
	SYSTEMConfigPerformance(FCY);

	init_pps();
	init_io();
	init_uart2();

	if ((memory = malloc(RAM_SIZE)) == NULL) {
		printf("\r\nsystem initialisation error\r\n");
		while (1)
			;
	}
	for (uint32_t i = 0; i < RAM_SIZE;) {
		poke(i++, 0x00);
		poke(i++, 0xFF);
	}

	/* Simulate PLUS/4 TEDMON */
	while (1)
		monitor();
}
