/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

/*
 * 6502 simulator
 */

#include "ape65.h"

uint8_t *memory = NULL;

extern const uint8_t BASIC[BASIC_SIZE];

extern const uint8_t MONITOR[MONITOR_SIZE];

uint8_t
peek(uint16_t address)
{
	if (address < RAM_SIZE) {
		return memory[address];
	}
	if (address < (BASIC_ADDR + BASIC_SIZE)) {
		return BASIC[address - BASIC_ADDR];
	}
	if (address >= MONITOR_ADDR) {
		return MONITOR[address - MONITOR_ADDR];
	}
	if (address == 0xF004) {
		return getcharacter();
	}
	if (address == 0xFD10) {
		return PORTE;
	}
	return 0xFF;
}

void
poke(uint16_t address, uint8_t value)
{
	if (address < RAM_SIZE) {
		memory[address] = value;
		return;
	}
	if (address == 0xF001) {
		putcharacter(value);
		return;
	}
	if (address == 0xFD10) {
		LATEbits.LATE7 = (value & 0x80) != 0;
		TRISE = value & 0x7F;
		return;
	}
}

static inline void
init_io(void)
{
	/* ADC Digital Mode */
	AD1PCFG = 0xFFFF;

	/* RESET E */
	LATE = 0x00;

	/* LED O/P */
	TRISE = 0x7F;
}

int
main(int argc, char **argv)
{
	SYSTEMConfigPerformance(FCY);

	/* Disable JTAG */
	DDPCONbits.JTAGEN = 0;

	init_io();
	init_uart2();

	if ((memory = malloc(RAM_SIZE)) == NULL) {
		printf("\r\nsystem initialisation error\r\n");
		while (1)
			;
	}
	for (uint32_t i = 0; i < RAM_SIZE;) {
		poke(i++, 0x00);
		poke(i++, 0xFF);
	}

	/* Simulate PLUS/4 TEDMON */
	while (1)
		monitor();
}
