/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

#include "6502.h"

extern struct register_file reg;

void
SBC(uint8_t op1, uint8_t op2)
{
	uint16_t acc;
	uint8_t carryflag = reg.SR & CARRY;

	op2 = ~op2;

	acc = op1 + op2 + carryflag;
	zero_negative(acc);
	overflow(acc, op1, op2);
	carry((acc & CARRYOUT) != 0);

	if (reg.SR & DECIMAL) {
		acc = (op1 & 0x0F) + (op2 & 0x0F) + carryflag;
		carryflag = acc & HALFCARRY;
		if (acc < 0x10)
		        acc += 0x0A;
		reg.AC = acc & 0x0F;
		acc = (op1 & 0xF0) + (op2 & 0xF0) + carryflag;
		if (acc < 0x100)
		        acc += 0xA0;
		reg.AC |= acc & 0xF0;
	} else {
		reg.AC = acc;
	}
}
