/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of Ape65.
 * 
 * Ape65 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * Ape65 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with Ape65. If not, see http://www.gnu.org/licenses/
 */

#include "ape65.h"

char *
skipspace(char **str)
{
	while (**str && **str <= ' ')
		(*str)++;

	return (*str);
}

uint8_t
toint(char c)
{
	if (c >= 'a' && c <= 'z')
		c = c - 'a' + 10;
	else if (c >= 'A' && c <= 'Z')
		c = c - 'A' + 10;
	else if (c >= '0' && c <= '9')
		c = c - '0';
	else
		c = 0;

	return (c);
}

uint16_t
strtoint(char **str, uint8_t base)
{
	uint16_t i = 0;

	skipspace(str);

	while ((**str >= 'a' && **str <= 'z') ||
		(**str >= 'A' && **str <= 'Z') ||
		(**str >= '0' && **str <= '9')) {
		i = i * base + toint(**str);
		(*str)++;
	}

	skipspace(str);

	return (i);
}

char *
rtrim(char *str, int16_t slen)
{
	uint16_t i = 0;

	str[slen - 1] = '\0';

	while (str[i] && str[i] >= ' ')
		++i;
	str[i]= '\0';

	return (str);
}

char
tochar(char c)
{
	c += '0';

	if (c > '9')
		c += 7;

	return (c);
}

char *
tostring(char *str, int16_t slen, uint16_t number, uint8_t base, uint8_t width)
{
	char temp[SLEN];
	uint8_t i = 0, j = 0;

	do {
		temp[i++] = tochar(number % base);
		number = number / base;
	}
	while (number);

	while (i < width)
		temp[i++] = '0';

	do {
		str[j++] = temp[--i];
	}
	while (i);

	str[j] = '\0';

	return (str);
}

#if defined(APE_POSIX)
char *
getfilename(char **str, int16_t slen, char *fname)
{
	uint16_t i = 0;
	char term;

	if (**str == '\0')
		return NULL;

	if (**str == '\'' || **str == '"')
		term = *(*str)++;
	else
		return NULL;

	while (**str && **str != term)
		fname[i++] = *(*str)++;
	fname[i] = '\0';

	if (**str) {
		(*str)++;
		skipspace(str);
	}

	return (fname);
}
#endif

/*
 * strcasestr implementation for cygwin
 */
const char *
mystrcasestr(const char *haystack, const char *needle)
{
	int i, j;
	int nlen = strlen(needle);
	int hlen = strlen(haystack) - nlen;

	for (i = 0; i <= hlen; i++) {
		for (j = 0; j < nlen; j++) {
			if (tolower((int)haystack[i + j]) != tolower((int)needle[j]))
				break;
		}
		if (j == nlen) {
			return &haystack[i];
		}
	}
	return NULL;
}

char *
getstring(char *str, int16_t slen)
{
	int16_t i = 0;
	uint8_t c;

	do {
		c = getcharacter();
		if (c == '\r') {
			putcharacter('\r');
			putcharacter('\n');
			str[i] = '\0';
		} else if (c == 0x08) {
			if (i) {
				putcharacter(0x08);
				putcharacter(' ');
				putcharacter(0x08);
				str[--i] = '\0';
			}
		} else if (c >= ' ' && c < 127) {
			putcharacter(c);
			str[i++] = c;
		}
	}
	while (c != '\r' && i < slen);

	return (str);
}
